@@ -20,7 +20,7 @@ import sys
 import os
 import textwrap
 
-VERSION = '3.3.2'
+VERSION = '3.4'
 ENCODING = 'utf-8'
 PY3 = sys.version_info >= (3,)
 PACKAGE_DIR = os.path.dirname(__file__)
@@ -32,17 +32,9 @@ def error(msg):
 def warn(msg):
     sys.stderr.write('WARNING: {0}\n'.format(msg))
 
-def usage(msg):
-    sys.stderr.write('Usage: {0}\n'.format(msg))
-    sys.exit(1)
-
 def info(msg):
     sys.stderr.write('{0}\n'.format(msg))
 
-def version():
-    write_line('SkoolKit {0}'.format(VERSION))
-    sys.exit(0)
-
 def show_package_dir():
     write_line(PACKAGE_DIR)
     sys.exit(0)
@@ -71,12 +63,6 @@ def parse_int(num_str, default=None):
     except ValueError:
         return default
 
-def parse_int_arg(num_str, option, name):
-    try:
-        return get_int_param(num_str)
-    except ValueError:
-        raise UsageError('{0} ({1}) must be an integer'.format(name, option))
-
 def get_chr(code):
     return chr(code) if PY3 else unichr(code).encode(ENCODING)
 
@@ -120,9 +106,6 @@ def read_bin_file(fname):
             raise SkoolKitError('{0}: file not found'.format(fname))
         raise
 
-class UsageError(Exception):
-    pass
-
 class SkoolKitError(Exception):
     pass
 
