@@ -61,13 +61,25 @@ class AsmWriter:
         self.desc_width = self.line_width - 2
 
         # Line terminator and indent
-        self.end = '\r\n' if crlf else '\n'
-        self.indent = '\t' if tab else ' ' * self.indent_width
+        if crlf:
+            self.end = '\r\n'
+        else:
+            self.end = '\n'
+        if tab:
+            self.indent = '\t'
+        else:
+            self.indent = ' ' * self.indent_width
 
         # Label suffix
-        self.label_suffix = ':' if self._get_int_property(properties, 'label-colons', 1) else ''
+        if self._get_int_property(properties, 'label-colons', 1):
+            self.label_suffix = ':'
+        else:
+            self.label_suffix = ''
 
-        self.table_writer = TableWriter(self.desc_width, 10)
+        min_col_width = self._get_int_property(properties, 'wrap-column-width-min', 10)
+        self.table_writer = TableWriter(self.desc_width, min_col_width)
+
+        self.handle_unsupported_macros = self._get_int_property(properties, 'handle-unsupported-macros', 0)
 
         self.snapshot = self.parser.snapshot
         self._snapshots = [(self.snapshot, '')]
@@ -93,7 +105,11 @@ class AsmWriter:
             first_instruction = entry.instructions[0]
             org = first_instruction.org
             if org:
-                self.write_line('{0}{1} {2}'.format(self.indent, 'org' if self.lower else 'ORG', org))
+                if self.lower:
+                    org_dir = 'org'
+                else:
+                    org_dir = 'ORG'
+                self.write_line('{0}{1} {2}'.format(self.indent, org_dir, org))
                 self.write_line('')
             self.entry = entry
             self.print_entry()
@@ -132,6 +148,12 @@ class AsmWriter:
         end, params, p_text = parse_params(text, index, default)
         return end, p_text
 
+    def _expand_unsupported_macro(self, text, index):
+        if self.handle_unsupported_macros:
+            end, params, p_text = parse_params(text, index, chars=',:;-x{}')
+            return end, ''
+        raise UnsupportedMacroError()
+
     def pop_snapshot(self):
         """Discard the current memory snapshot and replace it with the one that
         was most recently saved (by
@@ -182,7 +204,10 @@ class AsmWriter:
 
     def expand_chr(self, text, index):
         # #CHRnum or #CHR(num)
-        offset = 1 if text[index:].startswith('(') else 0
+        if text[index:].startswith('('):
+            offset = 1
+        else:
+            offset = 0
         end, num = parse_ints(text, index + offset, 1)
         return end + offset, get_chr(num)
 
@@ -200,7 +225,10 @@ class AsmWriter:
         # #EREFSaddr
         end, address = parse_ints(text, index, 1)
         ereferrers = self.parser.get_entry_point_refs(address)
-        addr_fmt = '{0:04X}' if text[index] == '$' else '{0}'
+        if text[index] == '$':
+            addr_fmt = '${0:04X}'
+        else:
+            addr_fmt = '{0}'
         if not ereferrers:
             raise MacroParsingError('Entry point at {0} has no referrers'.format(addr_fmt.format(address)))
         ereferrers.sort()
@@ -218,7 +246,7 @@ class AsmWriter:
         return self._expand_item_macro(text, index, 'fact')
 
     def expand_font(self, text, index):
-        raise UnsupportedMacroError()
+        return self._expand_unsupported_macro(text, index)
 
     def expand_html(self, text, index):
         # #HTML(text)
@@ -240,11 +268,6 @@ class AsmWriter:
             raise MacroParsingError("No link text specified: {0}{1}".format(macro, text[index:end]))
         return end, link_text
 
-    def expand_name(self, text, index):
-        # #NAMEname
-        end, params, p_text = parse_params(text, index)
-        return end, ''
-
     def expand_poke(self, text, index):
         # #POKE[#name][(link text)]
         return self._expand_item_macro(text, index, 'poke')
@@ -285,8 +308,10 @@ class AsmWriter:
             instructions = self.parser.instructions.get(address)
             if instructions:
                 label = instructions[0].addr_str
+            elif addr_str.startswith('$'):
+                label = addr_str[1:]
             else:
-                label = addr_str[1:] if addr_str.startswith('$') else addr_str
+                label = addr_str
         return end, label
 
     def expand_refs(self, text, index):
@@ -296,7 +321,10 @@ class AsmWriter:
         entry = self.entries.get(address)
         if not entry:
             raise MacroParsingError('No entry at {0}'.format(addr_str))
-        addr_fmt = '{0:04X}' if text[index] == '$' else '{0}'
+        if text[index] == '$':
+            addr_fmt = '${0:04X}'
+        else:
+            addr_fmt = '{0}'
         referrers = [ref.address for ref in entry.referrers]
         if referrers:
             referrers.sort()
@@ -318,22 +346,29 @@ class AsmWriter:
             raise MacroParsingError('Missing register argument')
         if len(reg) > 3 or any([char not in "abcdefhlirspxy'" for char in reg]):
             raise MacroParsingError('Bad register: "{0}"'.format(reg))
-        return end, reg.lower() if self.lower else reg.upper()
+        if self.lower:
+            reg_name = reg.lower()
+        else:
+            reg_name = reg.upper()
+        return end, reg_name
 
     def expand_scr(self, text, index):
-        raise UnsupportedMacroError()
+        return self._expand_unsupported_macro(text, index)
 
     def expand_space(self, text, index):
         # #SPACE[num] or #SPACE([num])
-        offset = 1 if text[index:].startswith('(') else 0
+        if text[index:].startswith('('):
+            offset = 1
+        else:
+            offset = 0
         end, num_sp = parse_ints(text, index + offset, 1, (1,))
         return end + offset, ''.ljust(num_sp)
 
     def expand_udg(self, text, index):
-        raise UnsupportedMacroError()
+        return self._expand_unsupported_macro(text, index)
 
     def expand_udgarray(self, text, index):
-        raise UnsupportedMacroError()
+        return self._expand_unsupported_macro(text, index)
 
     def expand(self, text):
         if text.find('#') < 0:
@@ -422,12 +457,7 @@ class AsmWriter:
             for line in lines:
                 if not ignoreua:
                     uaddress = self.find_unconverted_address(line)
-                    if uaddress and not (line.startswith('| ') and line.endswith(' |')):
-                        # Print a warning only if the unconverted address
-                        # appears outside a table row (there are many
-                        # unconvertible addresses in table rows that would
-                        # otherwise flood the warning output, so we ignore
-                        # them)
+                    if uaddress:
                         if instruction:
                             if not instruction.ignoremrcua:
                                 self.warn('Comment above {0} contains address ({1}) not converted to a label:\n; {2}'.format(instruction.address, uaddress, line))
@@ -458,7 +488,10 @@ class AsmWriter:
         instructions = self.entry.instructions
 
         while i < len(instructions) or lines:
-            instruction = instructions[i] if i < len(instructions) else None
+            if i < len(instructions):
+                instruction = instructions[i]
+            else:
+                instruction = None
 
             # Deal with remaining comment lines or rowspan on the previous
             # instruction
@@ -498,128 +531,111 @@ class TableWriter:
     def __init__(self, max_width, min_col_width):
         self.max_width = max_width
         self.min_col_width = min_col_width
-        self.cell_padding = len(' | ')
         self.table = None
         self.table_parser = TableParser()
+        self.cell_matrix = None
 
     def format_table(self, text):
         self.table = self.table_parser.parse_table(text)
-        self.table.cell_padding = self.cell_padding
+        self.table.cell_padding = 3
         self.table.prepare_cells()
         self.table.reduce_width(self.max_width, self.min_col_width)
+        self.cell_matrix = self._build_cell_matrix()
         return self._create_table_text()
 
+    def _build_cell_matrix(self):
+        cell_matrix = []
+        for row in self.table.rows:
+            cell_matrix.append([None] * self.table.num_cols)
+        for cell in self.table.cells:
+            for x in range(cell.col_index, cell.col_index + cell.colspan):
+                for y in range(cell.row_index, cell.row_index + cell.rowspan):
+                    cell_matrix[y][x] = cell
+        return cell_matrix
+
+    def _get_cell(self, col_index, row_index):
+        if 0 <= row_index < len(self.cell_matrix) and 0 <= col_index < self.table.num_cols:
+            return self.cell_matrix[row_index][col_index]
+
+    def _render_row(self, lines, row_index, first_line=True):
+        rendered = False
+        line = ''
+        col_index = 0
+        cell_left_transparent = True
+        cell = self._get_cell(col_index, row_index)
+        while cell:
+            if cell.transparent and cell_left_transparent:
+                border = ' '
+            else:
+                border = '|'
+            text = ''
+            if cell.contents and (first_line or row_index == cell.row_index + cell.rowspan - 1):
+                text = cell.contents.pop(0)
+                rendered = True
+            line += "{} {} ".format(border, text.ljust(self.table.get_cell_width(col_index, cell.colspan)))
+            col_index += cell.colspan
+            cell_left_transparent = cell.transparent
+            cell = self._get_cell(col_index, row_index)
+        if rendered:
+            if (cell and not cell.transparent) or (cell is None and not cell_left_transparent):
+                line += '|'
+            lines.append(line.rstrip())
+        return rendered
+
     def _create_table_text(self):
-        header_rows = [-1] + self.table.get_header_rows()
+        max_row_index = len(self.table.rows)
+        separator_row_indexes = set((0, max_row_index))
+        separator_row_indexes.update([i + 1 for i in self.table.get_header_rows()])
         lines = []
-        prev_row = [(None, 1, 0)] * self.table.num_cols
-        for row_index, row in enumerate(self.table.rows + [[]]):
-            # Deal with remaining lines in previous row
-            finished = row_index == 0
-            while not finished:
-                line = ''
-                adj_transparent = True
-                finished = True
-                col_index = 0
-                while col_index < self.table.num_cols:
-                    prev_cell, prev_rowspan, prev_line_index = prev_row[col_index]
-                    line += ' ' if adj_transparent and prev_cell.transparent else '|'
-                    if prev_cell:
-                        adj_transparent = prev_cell.transparent
-                        colspan = prev_cell.colspan
-                        contents = prev_cell.contents[prev_line_index] if prev_line_index < len(prev_cell.contents) else ''
-                    else:
-                        colspan = 1
-                        contents = ''
-                    if contents:
-                        finished = False
-                    line += " {0} ".format(contents.ljust(self.table.get_cell_width(col_index, colspan)))
-                    prev_row[col_index] = (prev_cell, prev_rowspan, prev_line_index + 1)
-                    col_index += colspan
-                if not finished:
-                    line += ' ' if prev_cell.transparent else '|'
-                    lines.append(line)
-
-            # Stop now if we've reached the end of the table (marked by an
-            # empty row)
-            if not row:
-                break
-
-            # Create a separator row if required
-            if row_index - 1 in header_rows:
-                lines.append(self._create_row_separator(row, fill=row_index > 0))
-
-            line = ''
-            adj_transparent = True
-            col_index = 0
-            for cell in row:
-                # Deal with previous rowspan > 1
-                prev_cell, prev_rowspan, prev_line_index = prev_row[col_index]
-                while prev_rowspan > 1:
-                    line += ' ' if adj_transparent and prev_cell.transparent else '|'
-                    adj_transparent, colspan = prev_cell.transparent, prev_cell.colspan
-                    contents = prev_cell.contents[prev_line_index] if prev_line_index < len(prev_cell.contents) else ''
-                    line += " {0} ".format(contents.ljust(self.table.get_cell_width(col_index, colspan)))
-                    prev_row[col_index] = (prev_cell, prev_rowspan - 1, prev_line_index + 1)
-                    col_index += colspan
-                    prev_cell, prev_rowspan, prev_line_index = prev_row[col_index]
-
-                # Deal with cells in this row
-                line += ' ' if adj_transparent and cell.transparent else '|'
-                adj_transparent, colspan = cell.transparent, cell.colspan
-                line += " {0} ".format(cell.contents[0].ljust(self.table.get_cell_width(col_index, colspan)))
-                prev_row[col_index] = (cell, cell.rowspan, 1)
-                col_index += colspan
-
-            # Deal with cells at the end of the previous row that have rowspan
-            # greater than 1
-            while col_index < len(prev_row):
-                prev_cell, prev_rowspan, prev_line_index = prev_row[col_index]
-                if prev_rowspan > 1:
-                    line += ' ' if adj_transparent and prev_cell.transparent else '|'
-                    adj_transparent, colspan = prev_cell.transparent, prev_cell.colspan
-                    contents = prev_cell.contents[prev_line_index] if prev_line_index < len(prev_cell.contents) else ''
-                    line += " {0} ".format(contents.ljust(self.table.get_cell_width(col_index, colspan)))
-                    prev_row[col_index] = (prev_cell, prev_rowspan - 1, prev_line_index + 1)
-                    col_index += prev_cell.colspan
-                else:
-                    break
-
-            line += ' ' if cell.transparent and row_index == 0 else '|'
-            lines.append(line)
-
-        lines.append(self._create_row_separator(self.table.rows[-1], prev_row))
+        for row_index in range(max_row_index):
+            if row_index in separator_row_indexes:
+                lines.append(self._create_row_separator(row_index))
+            self._render_row(lines, row_index)
+            while self._render_row(lines, row_index, False):
+                pass
+        lines.append(self._create_row_separator(max_row_index))
         return lines
 
-    def _create_row_separator(self, row, prev_row=None, fill=True):
-        separator = ''
+    def _create_row_separator(self, row_index):
+        # Return a separator between rows `row_index - 1` and `row_index`
+        line = ''
         col_index = 0
-        adj_transparent = True
-        for cell in row:
-            solid = fill or not cell.transparent
+        cell_left_contents = True
 
-            # Deal with previous rowspan > 1
-            if col_index < cell.col_index:
-                separator += '+' if fill else ' '
-                colspan = cell.col_index - col_index
-                spacer = '-' if solid else ' '
-                separator += spacer * (2 + self.table.get_cell_width(col_index, colspan))
+        while col_index < self.table.num_cols:
+            cell_above = self._get_cell(col_index, row_index - 1)
+            if row_index < len(self.table.rows):
+                cell = self._get_cell(col_index, row_index)
+            else:
+                cell = cell_above
+            if cell is None:
+                break
+            cell_above_transparent = cell_above is None or cell_above.transparent
+            cell_above_left = self._get_cell(col_index - 1, row_index - 1)
+            cell_above_left_transparent = cell_above_left is None or cell_above_left.transparent
+            cell_left = self._get_cell(col_index - 1, row_index)
+            cell_left_transparent = cell_left is None or cell_left.transparent
+            cell_contents = bool(cell_above and cell_above.contents)
+            if cell.transparent and cell_above_transparent and cell_above_left_transparent and cell_left_transparent:
+                line += ' '
+            elif cell_contents and cell_left_contents:
+                line += '|'
+            else:
+                line += '+'
+            if cell_contents:
+                text = cell.contents.pop(0)
+                line += ' {} '.format(text.ljust(self.table.get_cell_width(col_index, cell_above.colspan)))
+            else:
+                if cell.transparent and cell_above_transparent:
+                    spacer = ' '
+                else:
+                    spacer = '-'
+                line += spacer * (2 + self.table.get_cell_width(col_index, cell.colspan))
+            cell_left_contents = cell_contents
+            col_index += cell.colspan
 
-            # Deal with cells in this row
-            col_index, colspan = cell.col_index, cell.colspan
-            separator += ' ' if adj_transparent and not solid else '+'
-            spacer = '-' if solid else ' '
-            separator += spacer * (2 + self.table.get_cell_width(col_index, colspan))
-            adj_transparent = cell.transparent
-            col_index += colspan
-
-        # Deal with cells at the end of the previous row that have rowspan
-        # greater than 1
-        if prev_row:
-            while col_index < self.table.num_cols:
-                prev_cell = prev_row[col_index][0]
-                separator += ' ' if adj_transparent and not solid else '+'
-                separator += spacer * (2 + self.table.get_cell_width(col_index, prev_cell.colspan))
-                col_index += prev_cell.colspan
-
-        return separator + ('+' if solid else '')
+        if cell_contents:
+            return line + '|'
+        if line.endswith(' '):
+            return line.rstrip()
+        return line + '+'
