@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 
-# Copyright 2008-2012 Richard Dymond (rjdymond@gmail.com)
+# Copyright 2008-2013 Richard Dymond (rjdymond@gmail.com)
 #
 # This file is part of SkoolKit.
 #
@@ -103,8 +103,9 @@ class HtmlWriter:
         self.ref_parser = ref_parser
         self.file_info = file_info
         self.image_writer = image_writer
-        self.default_image_format = image_writer.default_format if image_writer else None
+        self.default_image_format = image_writer.default_format
         self.case = case
+        self.base = skool_parser.base
 
         self.snapshot = self.parser.snapshot
         self._snapshots = [(self.snapshot, '')]
@@ -170,6 +171,8 @@ class HtmlWriter:
         if def_game_name.endswith(suffix):
             def_game_name = def_game_name[:-len(suffix)]
         self.game = self.game_vars.get('Game', def_game_name)
+        link_operands = self.game_vars.get('LinkOperands', 'CALL,DEFW,DJNZ,JP,JR')
+        self.link_operands = tuple(op.upper() for op in link_operands.split(','))
 
         self.bugs = self.get_sections('Bug', True)
         self.facts = self.get_sections('Fact', True)
@@ -403,7 +406,9 @@ class HtmlWriter:
     def _get_changelog(self):
         changelog = []
         for title, paragraphs in self.get_sections('Changelog', True, True):
-            intro = '' if paragraphs[0][0] == '-' else paragraphs[0][0]
+            intro = paragraphs[0][0]
+            if intro == '-':
+                intro = ''
             items = []
             for p in paragraphs[1:]:
                 items.append(p)
@@ -740,6 +745,10 @@ class HtmlWriter:
     def write_entry(self, cwd, entry, fname=None, page_header=None, prev_entry=None, next_entry=None, up=None):
         fname = fname or FileInfo.asm_fname(entry.address)
         ofile = self.file_info.open_file(cwd, fname)
+        title_suffix = ''
+        asm_label = self.parser.get_asm_label(entry.address)
+        if asm_label:
+            title_suffix = ' ({0})'.format(asm_label)
         body_class = "disassembly"
         if entry.is_routine():
             title_format = "Routine at {0}"
@@ -753,7 +762,7 @@ class HtmlWriter:
         else:
             title_format = "Data at {0}"
             page_header = page_header or "Data"
-        self.write_header(ofile, title_format.format(entry.addr_str), cwd, body_class, page_header)
+        self.write_header(ofile, title_format.format(entry.addr_str) + title_suffix, cwd, body_class, page_header)
 
         prev_link = None
         if prev_entry:
@@ -770,7 +779,10 @@ class HtmlWriter:
         self.write_prev_next(ofile, prev_link, next_link, up_link)
 
         desc = self.expand(entry.description, cwd)
-        ofile.write('<div class="description">{0}: {1}</div>\n'.format(entry.addr_str, desc))
+        label_text = ''
+        if asm_label:
+            label_text = '{0}: '.format(asm_label)
+        ofile.write('<div class="description">{0}{1}: {2}</div>\n'.format(label_text, entry.addr_str, desc))
 
         table_class = 'disassembly'
         comment_class = 'comment'
@@ -781,8 +793,21 @@ class HtmlWriter:
             transparent_class = 'transparentDataComment'
         ofile.write('<table class="{0}">\n'.format(table_class))
 
+        show_comment_col = False
+        show_label_col = False
+        routine_comment_colspan = 3
+        for instruction in entry.instructions:
+            if instruction.comment and instruction.comment.text:
+                show_comment_col = True
+            if instruction.asm_label:
+                show_label_col = True
+                routine_comment_colspan = 4
+            if show_comment_col and show_label_col:
+                break
+        routine_comment_td = '<td class="routineComment" colspan="{}">\n'.format(routine_comment_colspan)
+
         ofile.write('<tr>\n')
-        ofile.write('<td class="routineComment" colspan="3">\n')
+        ofile.write(routine_comment_td)
         ofile.write('<div class="details">\n')
         if entry.details:
             ofile.write('{0}\n'.format(self.join_paragraphs(entry.details, cwd)))
@@ -792,42 +817,50 @@ class HtmlWriter:
         ofile.write('</td>\n')
         ofile.write('</tr>\n')
 
-        show_comment_col = False
-        for instruction in entry.instructions:
-            if instruction.comment and instruction.comment.text:
-                show_comment_col = True
         for instruction in entry.instructions:
             mid_routine_comments = entry.get_mid_routine_comment(instruction.label)
             address = instruction.address
             anchor = '<a name="{0}"></a>'.format(address)
             if mid_routine_comments:
                 ofile.write('<tr>\n')
-                ofile.write('<td class="routineComment" colspan="3">\n')
+                ofile.write(routine_comment_td)
                 ofile.write('{0}\n'.format(anchor))
                 ofile.write('<div class="comments">\n{0}\n</div>\n'.format(self.join_paragraphs(mid_routine_comments, cwd)))
                 ofile.write('</td>\n')
                 ofile.write('</tr>\n')
                 anchor = ''
             ofile.write('<tr>\n')
-            tdclass = 'label' if instruction.ctl in 'c*!' else 'address'
+            if show_label_col:
+                ofile.write('<td class="asmLabel">{}</td>\n'.format(instruction.asm_label or ''))
+            if instruction.ctl in 'c*!':
+                tdclass = 'label'
+            else:
+                tdclass = 'address'
             ofile.write('<td class="{0}">{1}{2}</td>\n'.format(tdclass, anchor, instruction.addr_str))
             operation, reference = instruction.operation, instruction.reference
-            if reference:
+            if reference and operation.upper().startswith(self.link_operands):
                 entry_address = reference.entry.address
-                name = '' if reference.address == entry_address else '#{0}'.format(reference.address)
+                if reference.address == entry_address:
+                    name = ''
+                else:
+                    name = '#{0}'.format(reference.address)
                 if reference.entry.asm_id:
                     # This is a reference to an entry in another disassembly
                     entry_file = FileInfo.asm_relpath(cwd, entry_address, self.get_code_path(reference.entry.asm_id))
                 else:
                     # This is a reference to an entry in the same disassembly
                     entry_file = FileInfo.asm_fname(entry_address)
-                link = '<a class="link" href="{0}{1}">{2}</a>'.format(entry_file, name, reference.addr_str)
+                link_text = self.parser.get_asm_label(reference.address) or reference.addr_str
+                link = '<a class="link" href="{0}{1}">{2}</a>'.format(entry_file, name, link_text)
                 operation = operation.replace(reference.addr_str, link)
             ofile.write('<td class="instruction">{0}</td>\n'.format(operation))
             if show_comment_col:
                 comment = instruction.comment
                 if comment:
-                    rowspan = '' if comment.rowspan < 2 else ' rowspan="{0}"'.format(comment.rowspan)
+                    if comment.rowspan < 2:
+                        rowspan = ''
+                    else:
+                        rowspan = ' rowspan="{0}"'.format(comment.rowspan)
                     comment_html = self.expand(comment.text, cwd)
                     ofile.write('<td class="{0}"{1}>{2}</td>\n'.format(comment_class, rowspan, comment_html))
             else:
@@ -836,7 +869,7 @@ class HtmlWriter:
 
         if entry.end_comment:
             ofile.write('<tr>\n')
-            ofile.write('<td class="routineComment" colspan="3">\n')
+            ofile.write(routine_comment_td)
             ofile.write('<div class="comments">\n{0}\n</div>\n'.format(self.join_paragraphs(entry.end_comment, cwd)))
             ofile.write('</td>\n')
             ofile.write('</tr>\n')
@@ -850,7 +883,10 @@ class HtmlWriter:
         prev_entry = None
         memory_map = self.memory_map
         for i, entry in enumerate(memory_map):
-            next_entry = memory_map[i + 1] if i + 1 < len(memory_map) else None
+            if i + 1 < len(memory_map):
+                next_entry = memory_map[i + 1]
+            else:
+                next_entry = None
             self.write_entry(cwd, entry, page_header=page_header, prev_entry=prev_entry, next_entry=next_entry, up=map_file)
             prev_entry = entry
 
@@ -911,7 +947,11 @@ class HtmlWriter:
                 desc_class = 'gbufferDesc'
             elif entry.ctl in 'uz':
                 size = entry.size
-                purpose = 'Unused ({0} byte{1})'.format(size, 's' if size > 1 else '')
+                if size > 1:
+                    suffix = 's'
+                else:
+                    suffix = ''
+                purpose = 'Unused ({0} byte{1})'.format(size, suffix)
                 entry_class = 'unused'
                 desc_class = 'unusedDesc'
             elif entry.ctl == 't':
@@ -951,6 +991,7 @@ class HtmlWriter:
     def write_head(self, ofile, title, cwd, js=False):
         ofile.write(PROLOGUE)
         ofile.write('<head>\n')
+        ofile.write('<meta http-equiv="content-type" content="text/html; charset=utf-8" />\n')
         ofile.write('<title>{0}: {1}</title>\n'.format(self.game, title))
         for css_file in self.paths['StyleSheet'].split(';'):
             ofile.write('<link rel="stylesheet" type="text/css" href="{0}" />\n'.format(FileInfo.relpath(cwd, join(self.paths['StyleSheetPath'], basename(css_file)))))
@@ -1028,13 +1069,22 @@ class HtmlWriter:
     def build_table(self, table):
         table_class = table.table_class
         html = []
-        class_attr = ' class="{0}"'.format(table_class) if table_class else ''
+        if table_class:
+            class_attr = ' class="{0}"'.format(table_class)
+        else:
+            class_attr = ''
         html.append('<table{0}>'.format(class_attr))
         for row in table.rows:
             html.append('<tr>')
             for cell in row:
-                colspan = ' colspan="{0}"'.format(cell.colspan) if cell.colspan > 1 else ''
-                rowspan = ' rowspan="{0}"'.format(cell.rowspan) if cell.rowspan > 1 else ''
+                if cell.colspan > 1:
+                    colspan = ' colspan="{0}"'.format(cell.colspan)
+                else:
+                    colspan = ''
+                if cell.rowspan > 1:
+                    rowspan = ' rowspan="{0}"'.format(cell.rowspan)
+                else:
+                    rowspan = ''
                 if cell.header:
                     tag = 'th'
                     td_class = ''
@@ -1044,7 +1094,10 @@ class HtmlWriter:
                     if cell.transparent:
                         cell_class += " transparent"
                     cell_class = cell_class.lstrip()
-                    td_class = ' class="{0}"'.format(cell_class) if cell_class else ''
+                    if cell_class:
+                        td_class = ' class="{0}"'.format(cell_class)
+                    else:
+                        td_class = ''
                 html.append('<{0}{1}{2}{3}>{4}</{0}>'.format(tag, td_class, colspan, rowspan, cell.contents))
             html.append('</tr>')
         html.append('</table>')
@@ -1052,7 +1105,10 @@ class HtmlWriter:
 
     def build_list(self, list_obj):
         html = []
-        class_attr = ' class="{0}"'.format(list_obj.css_class) if list_obj.css_class else ''
+        if list_obj.css_class:
+            class_attr = ' class="{0}"'.format(list_obj.css_class)
+        else:
+            class_attr = ''
         html.append('<ul{0}>'.format(class_attr))
         for item in list_obj.items:
             html.append('<li>{0}</li>'.format(item))
@@ -1217,7 +1273,10 @@ class HtmlWriter:
 
     def expand_chr(self, text, index, cwd):
         # #CHRnum or #CHR(num)
-        offset = 1 if text[index:].startswith('(') else 0
+        if text[index:].startswith('('):
+            offset = 1
+        else:
+            offset = 0
         end, num = parse_ints(text, index + offset, 1)
         return end + offset, '&#{0};'.format(num)
 
@@ -1235,18 +1294,16 @@ class HtmlWriter:
         # #EREFSaddr
         end, address = parse_ints(text, index, 1)
         ereferrers = self.get_entry_point_refs(address)
-        addr_fmt = '{0:04X}' if text[index] == '$' else '{0}'
         if not ereferrers:
-            raise MacroParsingError('Entry point at {0} has no referrers'.format(addr_fmt.format(address)))
-        ereferrers.sort()
-        eref_files = [(FileInfo.asm_relpath(cwd, addr, self.code_path), addr) for addr in ereferrers]
-        html = 'routine at '
-        if len(ereferrers) > 1:
+            raise MacroParsingError('Entry point at {0} has no referrers'.format(address))
+        if len(ereferrers) == 1:
+            html = 'routine at '
+        else:
+            ereferrers.sort()
             html = 'routines at '
-            html += ', '.join('<a class="link" href="{0}">{1}</a>'.format(eref_file, addr_fmt.format(addr)) for eref_file, addr in eref_files[:-1])
+            html += ', '.join('#R{}'.format(addr) for addr in ereferrers[:-1])
             html += ' and '
-        eref_file, addr = eref_files[-1]
-        html += '<a class="link" href="{0}">{1}</a>'.format(eref_file, addr_fmt.format(addr))
+        html += '#R{}'.format(ereferrers[-1])
         return end, html
 
     def expand_fact(self, text, index, cwd):
@@ -1298,11 +1355,6 @@ class HtmlWriter:
         end, list_obj = self.list_parser.parse_text(text, index)
         return end, self.build_list(list_obj)
 
-    def expand_name(self, text, index, cwd):
-        # #NAMEname
-        end, name, link_text = parse_params(text, index)
-        return end, '<a name="{0}"></a>'.format(name)
-
     def expand_poke(self, text, index, cwd):
         # #POKE[#name][(link text)]
         return self._expand_item_macro('#POKE', text, index, cwd, self.pokes, P_POKES, 'poke')
@@ -1353,9 +1405,11 @@ class HtmlWriter:
             inst_addr_str = self.get_instruction_addr_str(address)
             container_address = container.address
             code_path = self.code_path
-            anchor = anchor if address == container_address else '#{0}'.format(address)
+            if address != container_address:
+                anchor = '#{0}'.format(address)
+        asm_label = self.parser.get_asm_label(address)
         ref_file = FileInfo.asm_relpath(cwd, container_address, code_path)
-        link = '<a class="link" href="{0}{1}">{2}</a>'.format(ref_file, anchor, link_text or inst_addr_str)
+        link = '<a class="link" href="{0}{1}">{2}</a>'.format(ref_file, anchor, link_text or asm_label or inst_addr_str)
         return end, link
 
     def expand_refs(self, text, index, cwd):
@@ -1365,18 +1419,16 @@ class HtmlWriter:
         entry = self.entries.get(address)
         if not entry:
             raise MacroParsingError('No entry at {0}'.format(addr_str))
-        addr_fmt = '{0:04X}' if text[index] == '$' else '{0}'
         referrers = [ref.address for ref in entry.referrers]
         if referrers:
-            referrers.sort()
-            ref_files = [(FileInfo.asm_relpath(cwd, addr, self.code_path), addr) for addr in referrers]
-            html = ('{0} routine at '.format(prefix)).lstrip()
-            if len(referrers) > 1:
+            if len(referrers) == 1:
+                html = ('{0} routine at '.format(prefix)).lstrip()
+            else:
+                referrers.sort()
                 html = ('{0} routines at '.format(prefix)).lstrip()
-                html += ', '.join('<a class="link" href="{0}">{1}</a>'.format(ref_file, addr_fmt.format(addr)) for ref_file, addr in ref_files[:-1])
+                html += ', '.join('#R{}'.format(addr) for addr in referrers[:-1])
                 html += ' and '
-            ref_file, addr = ref_files[-1]
-            html += '<a class="link" href="{0}">{1}</a>'.format(ref_file, addr_fmt.format(addr))
+            html += '#R{}'.format(referrers[-1])
         else:
             html = 'Not used directly by any other routines'
         return end, html
@@ -1391,7 +1443,10 @@ class HtmlWriter:
             raise MacroParsingError('Missing register argument')
         if len(reg) > 3:
             raise MacroParsingError('Bad register: "{0}"'.format(reg))
-        reg_name = reg.lower() if self.case == CASE_LOWER else reg.upper()
+        if self.case == CASE_LOWER:
+            reg_name = reg.lower()
+        else:
+            reg_name = reg.upper()
         return end, '<span class="register">{0}</span>'.format(reg_name)
 
     def expand_scr(self, text, index, cwd):
@@ -1403,7 +1458,10 @@ class HtmlWriter:
 
     def expand_space(self, text, index, cwd):
         # #SPACE[num] or #SPACE([num])
-        offset = 1 if text[index:].startswith('(') else 0
+        if text[index:].startswith('('):
+            offset = 1
+        else:
+            offset = 0
         end, num_sp = parse_ints(text, index + offset, 1, (1,))
         return end + offset, '&#160;' * num_sp
 
@@ -1495,9 +1553,7 @@ class HtmlWriter:
                 raise SkoolParsingError('Found unsupported macro: {0}'.format(marker))
             except MacroParsingError as e:
                 raise SkoolParsingError('Error while parsing {0} macro: {1}'.format(marker, e.args[0]))
-            prefix = text[:start].rstrip() if rep.startswith('\n') else text[:start]
-            suffix = text[end:].lstrip() if rep.endswith('\n') else text[end:]
-            text = "{0}{1}{2}".format(prefix, rep, suffix)
+            text = text[:start] + rep + text[end:]
 
         return text
 
@@ -1510,6 +1566,7 @@ class FileInfo:
     :param replace_images: Whether existing images should be overwritten.
     """
     def __init__(self, topdir, game_dir, replace_images):
+        self.game_dir = game_dir
         self.odir = join(topdir, game_dir)
         self.replace_images = replace_images
 
@@ -1570,7 +1627,8 @@ class Udg(object):
             rbyte = 0
             for byte in tile_data:
                 rbyte //= 2
-                rbyte += 128 if byte & b else 0
+                if byte & b:
+                    rbyte += 128
             rotated.append(rbyte)
             b *= 2
         rotated.reverse()
