@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 
-# Copyright 2008-2012 Richard Dymond (rjdymond@gmail.com)
+# Copyright 2008-2013 Richard Dymond (rjdymond@gmail.com)
 #
 # This file is part of SkoolKit.
 #
@@ -19,7 +19,7 @@
 import cgi
 import re
 
-from . import warn, wrap, parse_int, open_file, SkoolParsingError
+from . import warn, wrap, get_int_param, parse_int, open_file, SkoolParsingError
 
 DIRECTIVES = 'bcgituwz'
 
@@ -46,12 +46,18 @@ BASE_10 = 10
 BASE_16 = 16
 
 def get_address(operation, check_prefix=True):
-    prefixes = '[ ,(+-]' if check_prefix else ''
+    if check_prefix:
+        prefixes = '[ ,(+-]'
+    else:
+        prefixes = ''
     search = re.search('{0}\$[0-9A-Fa-f]+'.format(prefixes), operation)
     if not search:
         search = re.search('{0}[0-9]+'.format(prefixes), operation)
     if search:
-        index = 1 if check_prefix else 0
+        if check_prefix:
+            index = 1
+        else:
+            index = 0
         return search.group()[index:]
 
 def get_instruction_ctl(op):
@@ -67,8 +73,12 @@ def get_instruction_ctl(op):
     return 'C'
 
 def get_defb_length(item_str, defb=True):
-    byte_fmt = '{0}' if defb else 'B{0}'
-    text_fmt = 'T{0}' if defb else '{0}'
+    if defb:
+        byte_fmt = '{0}'
+        text_fmt = 'T{0}'
+    else:
+        byte_fmt = 'B{0}'
+        text_fmt = '{0}'
     full_length = 0
     lengths = []
     length = 0
@@ -149,7 +159,10 @@ def set_bytes(snapshot, address, operation):
     args = [parse_int(v, 0) for v in operation[5:].split(',')]
     if directive == 'DEFS':
         span = args[0]
-        value = args[1] if len(args) > 1 else 0
+        if len(args) > 1:
+            value = args[1]
+        else:
+            value = 0
         snapshot[address:address + span] = [value] * span
     elif directive == 'DEFW':
         for arg in args:
@@ -161,7 +174,10 @@ def parse_asm_block_directive(directive, stack):
     infix = directive[len(prefix):len(prefix) + 1]
     suffix = directive[len(prefix) + len(infix):].rstrip()
     if prefix in ('ofix', 'bfix', 'rfix', 'isub', 'rsub') and infix in '+-' and suffix in ('begin', 'else', 'end'):
-        cur_op = stack[-1] if stack else (None, None)
+        if stack:
+            cur_op = stack[-1]
+        else:
+            cur_op = (None, None)
         if suffix == 'begin':
             if prefix == cur_op[0]:
                 raise SkoolParsingError('{0} inside {1}{2} block'.format(directive, cur_op[0], cur_op[1]))
@@ -201,10 +217,12 @@ class SkoolParser:
     :param html: Whether to escape HTML characters.
     :param create_labels: Whether to create default labels for unlabelled
                           instructions.
+    :param asm_labels: Whether to parse `@label` directives.
     """
-    def __init__(self, skoolfile, case=None, base=None, asm_mode=0, warnings=False, fix_mode=0, html=False, create_labels=False):
+    def __init__(self, skoolfile, case=None, base=None, asm_mode=0, warnings=False, fix_mode=0, html=False, create_labels=False, asm_labels=True):
         self.skoolfile = skoolfile
-        self.mode = Mode(case, base, asm_mode, warnings, fix_mode, html, create_labels)
+        self.mode = Mode(case, base, asm_mode, warnings, fix_mode, html, create_labels, asm_labels)
+        self.base = base
 
         self.snapshot = [0] * 65536  # 64K of Spectrum memory
         self.instructions = {}       # address -> [Instructions]
@@ -224,17 +242,26 @@ class SkoolParser:
         """Return the routine or data block that starts at `address`."""
         return self.entries.get(address)
 
+    def get_instruction(self, address):
+        """Return the instruction at `address`."""
+        return self.instructions.get(address, [None])[0]
+
     def get_container(self, address):
         """Return the routine or data block that contains `address`."""
-        instructions = self.instructions.get(address)
-        if instructions:
-            return instructions[0].container
+        instruction = self.get_instruction(address)
+        if instruction:
+            return instruction.container
 
     def get_entry_point_refs(self, address):
         """Return the addresses of the routines and data blocks that contain
         instructions that refer to `address`.
         """
-        return [referrer.address for referrer in self.instructions[address][0].referrers]
+        return [referrer.address for referrer in self.get_instruction(address).referrers]
+
+    def get_asm_label(self, address):
+        instruction = self.get_instruction(address)
+        if instruction:
+            return instruction.asm_label
 
     def _parse_skool(self, skoolfile):
         map_entry = None
@@ -317,6 +344,8 @@ class SkoolParser:
         # Do some post-processing
         self._parse_address_comments(address_comments)
         self._calculate_references()
+        if self.mode.asm_labels:
+            self._generate_labels()
         if self.mode.html:
             self._calculate_entry_sizes()
             self._escape_instructions()
@@ -340,7 +369,9 @@ class SkoolParser:
             paragraphs = [self._html_escape(p) for p in paragraphs]
         if split:
             return paragraphs
-        return paragraphs[0] if paragraphs else ''
+        if paragraphs:
+            return paragraphs[0]
+        return ''
 
     def _html_escape(self, text):
         chunks = []
@@ -391,13 +422,24 @@ class SkoolParser:
                         do_op = self.mode.asm
                     elif p == 'rsub':
                         do_op = self.mode.do_rsubs
-                    include = i == '+' if do_op else i == '-'
+                    if do_op:
+                        include = i == '+'
+                    else:
+                        include = i == '-'
                     if not include:
                         break
                 self.mode.include = include
                 return
 
-            if self.mode.asm and self.mode.include:
+            if not self.mode.include:
+                return
+
+            if directive.startswith('label='):
+                self.mode.label = directive[6:].rstrip()
+            elif directive.startswith('keep'):
+                self.mode.keep = True
+
+            if self.mode.asm:
                 if directive.startswith('rsub='):
                     self.mode.rsub = directive[5:].rstrip()
                 elif directive.startswith('ssub='):
@@ -408,8 +450,6 @@ class SkoolParser:
                     self.mode.bfix = directive[5:].rstrip()
                 elif directive.startswith('ofix='):
                     self.mode.ofix = directive[5:].rstrip()
-                elif directive.startswith('label='):
-                    self.mode.label = directive[6:].rstrip()
                 elif directive.startswith('nolabel'):
                     self.mode.nolabel = True
                 elif directive.startswith('nowarn'):
@@ -417,8 +457,6 @@ class SkoolParser:
                 elif directive.startswith('ignoreua'):
                     self.mode.ignoreua = True
                     self.ignores.append(len(self.comments))
-                elif directive.startswith('keep'):
-                    self.mode.keep = True
                 elif directive.startswith('org='):
                     self.mode.org = directive[4:].rstrip()
                 elif directive.startswith('writer='):
@@ -481,7 +519,10 @@ class SkoolParser:
             prefix = ''
             elements = line.split(None, 1)
             reg = elements[0]
-            desc = elements[1] if len(elements) > 1 else ''
+            if len(elements) > 1:
+                desc = elements[1]
+            else:
+                desc = ''
             if ':' in reg:
                 prefix, reg = reg.split(':', 1)
             if self.mode.lower:
@@ -522,19 +563,21 @@ class SkoolParser:
         # Parse operations for routine/data addresses
         for entry in self.memory_map:
             for instruction in entry.instructions:
+                if instruction.keep:
+                    continue
                 operation = instruction.operation.upper()
-                if operation.startswith('DEFW') or operation[0:2] in ['JP', 'JR', 'CA']:
+                if operation.startswith(('CALL', 'DEFW', 'DJNZ', 'JP', 'JR', 'LD ')):
                     addr_str = get_address(operation)
                     if not addr_str:
                         continue
                     address = parse_int(addr_str)
-                    other_instructions = self.instructions.get(address, [])
+                    other_instructions = self.instructions.get(address)
                     if other_instructions:
                         other_entry = other_instructions[0].container
-                        if entry != other_entry and (other_entry.is_remote() or operation.startswith('DEFW') or other_entry.is_routine()):
+                        if (entry != other_entry or self.mode.asm_labels) and (other_entry.is_remote() or operation.startswith(('DEFW', 'LD ')) or other_entry.is_routine()):
                             instruction.set_reference(other_entry, address, addr_str)
-                            for other_instruction in other_instructions:
-                                other_instruction.add_referrer(entry)
+                            if operation.startswith(('CALL', 'DJNZ', 'JP', 'JR')):
+                                other_instructions[0].add_referrer(entry)
 
     def _escape_instructions(self):
         for entry in self.memory_map:
@@ -546,7 +589,6 @@ class SkoolParser:
             warn(s)
 
     def _substitute_labels(self):
-        self._generate_labels()
         for entry in self.memory_map:
             for instruction in entry.instructions:
                 if instruction.sub or not instruction.keep:
@@ -574,8 +616,8 @@ class SkoolParser:
         operand = get_address(operation)
         if operand is None:
             return
-        operand_int = parse_int(operand, -1)
-        instructions = None if operand_int < 0 else self.instructions.get(operand_int)
+        operand_int = get_int_param(operand)
+        instructions = self.instructions.get(operand_int)
         if instructions:
             reference = instructions[0]
             if reference.asm_label:
@@ -597,7 +639,7 @@ class SkoolParser:
             self.warn('Unreplaced operand: {0} {1}'.format(instruction.address, operation))
 
 class Mode:
-    def __init__(self, case, base, asm_mode, warnings, fix_mode, html, create_labels):
+    def __init__(self, case, base, asm_mode, warnings, fix_mode, html, create_labels, asm_labels):
         self.lower = case == CASE_LOWER
         self.upper = case == CASE_UPPER
         self.decimal = base == BASE_10
@@ -614,9 +656,13 @@ class Mode:
         self.do_bfixes = fix_mode >= 2
         self.labels = []
         self.create_labels = create_labels
+        self.asm_labels = asm_labels
         self.ignoretua = False
         self.ignoredua = False
-        self.hex4fmt = '${0:04x}' if self.lower else '${0:04X}'
+        if self.lower:
+            self.hex4fmt = '${0:04x}'
+        else:
+            self.hex4fmt = '${0:04X}'
         self.reset()
 
     def reset(self):
@@ -643,35 +689,34 @@ class Mode:
             self.include = False
 
     def apply_asm_attributes(self, instruction):
-        if self.html:
-            return
+        instruction.keep = self.keep
 
-        if self.label:
+        if self.asm_labels and self.label:
             if self.label in self.labels:
                 raise SkoolParsingError('Duplicate label {0} at {1}'.format(self.label, instruction.address))
             self.labels.append(self.label)
             instruction.asm_label = self.label
 
-        sub = self.isub
-        if self.bfix is not None and self.do_bfixes:
-            sub = self.bfix
-        elif self.ofix is not None and self.do_ofixes:
-            sub = self.ofix
-        elif self.rsub is not None and self.do_rsubs:
-            sub = self.rsub
-        elif self.ssub is not None and self.do_ssubs:
-            sub = self.ssub
-        if sub is not None:
-            _, sub = self.apply_case('', sub)
-            _, sub = self.apply_base('', sub)
-            instruction.apply_sub(sub)
+        if not self.html:
+            sub = self.isub
+            if self.bfix is not None and self.do_bfixes:
+                sub = self.bfix
+            elif self.ofix is not None and self.do_ofixes:
+                sub = self.ofix
+            elif self.rsub is not None and self.do_rsubs:
+                sub = self.rsub
+            elif self.ssub is not None and self.do_ssubs:
+                sub = self.ssub
+            if sub is not None:
+                _, sub = self.apply_case('', sub)
+                _, sub = self.apply_base('', sub)
+                instruction.apply_sub(sub)
 
-        instruction.warn = not self.nowarn
-        instruction.nolabel = self.nolabel
-        instruction.ignoreua = self.ignoreua
-        instruction.ignoremrcua = self.ignoremrcua
-        instruction.keep = self.keep
-        instruction.org = self.org
+            instruction.warn = not self.nowarn
+            instruction.nolabel = self.nolabel
+            instruction.ignoreua = self.ignoreua
+            instruction.ignoremrcua = self.ignoremrcua
+            instruction.org = self.org
 
         self.reset()
 
@@ -815,7 +860,10 @@ class Instruction:
     def __init__(self, label, address, operation):
         self.label = label
         self.ctl = label[0]
-        self.addr_str = label[1:] if label[1].isdigit() else label[2:]
+        if label[1].isdigit():
+            self.addr_str = label[1:]
+        else:
+            self.addr_str = label[2:]
         self.address = address
         self.operation = operation
         self.container = None
@@ -823,10 +871,15 @@ class Instruction:
         self.comment = None
         self.referrers = []
         self.asm_label = None
+        self.nolabel = False
+        self.org = None
         # If this instruction has no address, it was inserted between
         # @rsub+begin and @rsub+end; in that case, mark it as a subbed
         # instruction already
-        self.sub = None if address else operation
+        if address:
+            self.sub = None
+        else:
+            self.sub = operation
         self.keep = False
         self.warn = True
         self.ignoreua = False
@@ -943,7 +996,7 @@ class TableParser:
         for ws_char in '\n\r\t':
             text = text.replace(ws_char, ' ')
 
-        index = len(TABLE_MARKER) if text.startswith(TABLE_MARKER) else 0
+        index = 0
         classes = []
         if text[index] == '(':
             end = text.find(')', index)
@@ -951,7 +1004,10 @@ class TableParser:
                 raise SkoolParsingError("Cannot find closing ')' in table CSS class list:\n{0}".format(table_def))
             classes = [c.strip() for c in text[index + 1:end].split(',')]
             index = end + 1
-        table_class = classes[0] if classes else ''
+        if classes:
+            table_class = classes[0]
+        else:
+            table_class = ''
         column_classes = classes[1:]
         wrap_columns = []
         for i, column_class in enumerate(column_classes):
@@ -978,7 +1034,10 @@ class TableParser:
                     prev_rowspan, prev_colspan = prev_spans.get(col_index, (1, 1))
                 header, transparent = False, False
                 rowspan, colspan = 1, 1
-                cell_class = column_classes[col_index] if len(column_classes) > col_index else ''
+                if len(column_classes) > col_index:
+                    cell_class = column_classes[col_index]
+                else:
+                    cell_class = ''
                 cell = cell.strip()
                 if cell.startswith('='):
                     end = cell.find(' ')
@@ -1133,7 +1192,7 @@ class ListParser:
         for ws_char in '\n\r\t':
             text = text.replace(ws_char, ' ')
 
-        index = len(LIST_MARKER) if text.startswith(LIST_MARKER) else 0
+        index = 0
         css_class = ''
         if text[index] == '(':
             end = text.find(')', index)
